/**
 *  \file ArrayVector.h
 *  std::vector ݊, Ce[^Ƃă|C^gpRei.
 */
#pragma once

#ifdef _MSC_VER
  #pragma warning( push )
  #pragma warning( disable : 4996 )
#endif

#include <memory>
#include <algorithm>
#include <iterator>
#include <type_traits>

//
// algorithm destroy_array:
// The reverse of std::unitialized_copy, takes a block of
// initialized memory and calls destructors on all objects therein.
//
// refer http://www.boost.org/libs/type_traits/examples/trivial_destructor_example.cpp
//
namespace detail {
//----------------------------------------------------------------
template <class T>
void do_destroy_array(T* first, T* last, const std::tr1::false_type&)
{
   while(first != last)
   {
      first->~T();
      ++first;
   }
}

template <class T>
inline void do_destroy_array(T* first, T* last, const std::tr1::true_type&)
{
}

template <class T>
inline void destroy_n(T* first, unsigned int size)
{
   detail::do_destroy_array(first, first + size, std::tr1::has_trivial_destructor<T>());
}

//----------------------------------------------------------------
} // namespace detail

//----------------------------------------------------------------
/**
 *  Ce[^|C^ł邱Ƃۏ؂ \a std::vector ݊̃Rei.
 */
template <typename T, typename Alloc = std::allocator<T> >
class ArrayVector
{
public:
  typedef T value_type;
  typedef value_type & reference;
  typedef value_type const & const_reference;
  typedef value_type * pointer;
  typedef value_type const * const_pointer;
  typedef value_type * iterator;
  typedef value_type const * const_iterator;
  typedef unsigned int size_type;
  typedef int difference_type;
  typedef Alloc allocator_type;
  typedef std::reverse_iterator<iterator> reverse_iterator;
  typedef std::reverse_iterator<const_iterator> const_reverse_iterator;

private:
  typedef ArrayVector this_type;
  enum { minimum_capacity = 2 };

public:
  // Constructors
  ArrayVector();
  explicit ArrayVector(Alloc const &alloc);
  explicit ArrayVector(size_type size, Alloc const &alloc = Alloc());
  ArrayVector(size_type size, value_type const & initial, Alloc const &alloc = Alloc());
  ArrayVector(this_type const &rhs);
  
  template <typename InputIterator>
  ArrayVector(InputIterator begin, InputIterator end);
  
  template <typename InputIterator>
  ArrayVector(InputIterator begin, InputIterator end, Alloc const &alloc);
  
  // Assign Operator
  this_type & operator = (this_type const &rhs);
  
  // Destructor
  ~ArrayVector();
  
  // Methods that return Iterator.
  const_iterator begin() const
  {
    return data_;
  }
  
  iterator begin()
  {
    return data_;
  }
  
  const_iterator end() const
  {
    return data_ + size_;
  }
  
  iterator end()
  {
    return data_ + size_;
  }
  
  const_reverse_iterator rbegin() const
  {
    return const_reverse_iterator(end());
  }

  reverse_iterator rbegin()
  {
    return reverse_iterator(end());
  }  
  
  const_reverse_iterator rend() const
  {
    return const_reverse_iterator(begin());
  }
  
  reverse_iterator rend()
  {
    return reverse_iterator(begin());
  }
  
  // Accessors
  const_iterator data() const
  {
    return data_;
  }
  
  iterator data()
  {
    return data_;
  }
  
  reference front()
  {
    return *data_;
  }
  
  const_reference front() const
  {
    return *data_;
  }
  
  reference back()
  {
    return data_[size_ - 1];
  }
  
  const_reference back() const
  {
    return data_[size_ - 1];
  }
  
  reference operator [] (size_type i)
  {
    return data_[i];
  }
  
  const_reference operator [] (size_type i) const
  {
    return data_[i];
  }
  
  // Methods that modify data.
  void pop_back();
  void push_back(value_type const &value);
  iterator insert(iterator p, value_type const &value);
  iterator insert(iterator p, size_type n, value_type const &value);
  
  template <typename InputIterator>
  iterator insert(iterator p, InputIterator begin, InputIterator end);
  
  iterator erase(iterator p);
  iterator erase(iterator p, iterator q);
  void clear();
  void reserve(size_type new_capacity);
  void reserve();
  void resize(size_type new_size, value_type const &initial);
  
  void resize(size_type new_size)
  {
    resize(new_size, value_type());
  }
  
  // Methods that inspect the size of data.
  bool empty() const
  {
    return size_ == 0;
  }
  
  size_type size() const
  {
    return size_;
  }
  
  size_type capacity() const
  {
    return capacity_;
  }
  
  // Swap Method
  void swap(this_type &rhs);

private:
  void deallocate(pointer data, size_type size);
  pointer reserve_raw(size_type capacity);

private:
  Alloc alloc_;
  size_type size_, capacity_;
  pointer data_;
};

//----------------------------------------------------------------
// Constructors
template <typename T, typename Alloc>
ArrayVector<T, Alloc>::ArrayVector()
  : alloc_(Alloc()),
    size_(0),
    capacity_(minimum_capacity),
    data_(reserve_raw(minimum_capacity))
{
}

template <typename T, typename Alloc>
ArrayVector<T, Alloc>::ArrayVector(Alloc const &alloc)
  : alloc_(alloc),
    size_(0),
    capacity_(minimum_capacity),
    data_(reserve_raw(minimum_capacity))
{
}

template <typename T, typename Alloc>
ArrayVector<T, Alloc>::ArrayVector(size_type size, Alloc const &alloc)
  : alloc_(alloc),
    size_(size),
    capacity_(size),
    data_(reserve_raw(size))
{
  if (size_ > 0)
  {
    std::uninitialized_fill(data_, data_ + size_, value_type());
  }
}

template <typename T, typename Alloc>
ArrayVector<T, Alloc>::ArrayVector(size_type size, value_type const & initial, Alloc const &alloc)
  : alloc_(alloc),
    size_(size),
    capacity_(size),
    data_(reserve_raw(size))
{
  if (size_ > 0)
  {
    std::uninitialized_fill(data_, data_ + size_, initial);
  }
}

template <typename T, typename Alloc>
ArrayVector<T, Alloc>::ArrayVector(this_type const &rhs)
  : alloc_(rhs.alloc_),
    size_(rhs.size_),
    capacity_(rhs.capacity_),
    data_(reserve_raw(rhs.capacity_))
{
  if (size_ > 0)
  {
    std::uninitialized_copy(rhs.data_, rhs.data_ + size_, data_);
  }
}

template <typename T, typename Alloc>
template <typename InputIterator>
ArrayVector<T, Alloc>::ArrayVector(InputIterator i, InputIterator iend)
  : alloc_(),
    size_(static_cast<size_type>(std::distance(i, iend))),
    capacity_(size_),
    data_(reserve_raw(size_))
{
  if (size_ > 0)
  {
    std::uninitialized_copy(i, iend, data_);
  }
}

template <typename T, typename Alloc>
template <typename InputIterator>
ArrayVector<T, Alloc>::ArrayVector(InputIterator i, InputIterator iend, Alloc const &alloc)
  : alloc_(alloc),
    size_(std::distance(i, iend)),
    capacity_(size_),
    data_(reserve_raw(size_))
{
  if (size_ > 0)
  {
    std::uninitialized_copy(i, iend, data_);
  }
}

// Assign Operator
template <typename T, typename Alloc>
typename ArrayVector<T, Alloc>::this_type & ArrayVector<T, Alloc>::operator = (this_type const &rhs)
{
  if (this == &rhs) return *this;
  ArrayVector new_vector(rhs);
  swap(new_vector);
  return *this;
}

// Destructor
template <typename T, typename Alloc>
ArrayVector<T, Alloc>::~ArrayVector()
{
  deallocate(data_, size_);
}

// Methods that modify data.
template <typename T, typename Alloc>
void ArrayVector<T, Alloc>::pop_back()
{
  --size_;
  alloc_.destroy(data_ + size_);
}

template <typename T, typename Alloc>
void ArrayVector<T, Alloc>::push_back(value_type const &t)
{
  reserve();
  alloc_.construct(data_ + size_, t);
  ++size_;
}

template <typename T, typename Alloc>
typename ArrayVector<T, Alloc>::iterator
  ArrayVector<T, Alloc>::insert(iterator p, value_type const &value)
{
  difference_type pos = p - begin();

  if (p == end())
  {
    push_back(v);
    p = begin() + pos;
  }
  else
  {
    push_back(back());
    p = begin() + pos;
    std::copy_backward(p, end() - 2, end() - 1);
    *p = v;
  }
  return p;
}

template <typename T, typename Alloc>
typename ArrayVector<T, Alloc>::iterator
  ArrayVector<T, Alloc>::insert(iterator p, size_type n, value_type const &value)
{
  difference_type pos = p - begin();
  size_type new_size = size() + n;
  
  if (new_size >= capacity_)
  {
    pointer new_data = reserve_raw(new_size);
    std::uninitialized_copy(begin(), p, new_data);
    std::uninitialized_fill(new_data + pos, new_data + pos + n, value);
    std::uninitialized_copy(p, end(), new_data + pos + n);
    deallocate(data_, size_);
    capacity_ = new_size;
    data_ = new_data;
  }
  else if (pos + n >= size_)
  {
    size_type diff = pos + n - size_;
    std::uninitialized_copy(p, end(), end() + diff);
    std::uninitialized_fill(end(), end() + diff, value);
    std::fill(p, end(), value);
  }
  else
  {
    size_type diff = size_ - (pos + n);
    std::uninitialized_copy(end() - n, end(), end());
    std::copy_backward(p, p + diff, end());
    std::fill(p, p + n, value);
  }
  
  size_ = new_size;
  return begin() + pos;
}

template <typename T, typename Alloc>
template <typename InputIterator>
typename ArrayVector<T, Alloc>::iterator
  ArrayVector<T, Alloc>::insert(iterator p, InputIterator i, InputIterator iend)
{
  size_type n = iend - i;
  size_type pos = p - begin();
  size_type new_size = size() + n;
  
  if (new_size >= capacity_)
  {
    pointer new_data = reserve_raw(new_size);
    std::uninitialized_copy(begin(), p, new_data);
    std::uninitialized_copy(i, iend, new_data + pos);
    std::uninitialized_copy(p, end(), new_data + pos + n);
    deallocate(data_, size_);
    capacity_ = new_size;
    data_ = new_data;
  }
  else if (pos + n >= size_)
  {
    size_type diff = pos + n - size;
    std::uninitialized_copy(p, end(), end() + diff);
    std::uninitialized_copy(iend - diff, iend, end());
    std::copy(i, iend - diff, p);
  }
  else
  {
    size_type diff = size_ - (pos + n);
    std::uninitialized_copy(end() - n, end(), end());
    std::copy_backward(p, p + diff, end());
    std::copy(i, iend, p);
  }
  
  size_ = new_size;
  return begin() + pos;
}

template <typename T, typename Alloc>
typename ArrayVector<T, Alloc>::iterator ArrayVector<T, Alloc>::erase(iterator p)
{
  std::copy(p + 1, end(), p);
  pop_back();
  return p;
}

template <typename T, typename Alloc>
typename ArrayVector<T, Alloc>::iterator ArrayVector<T, Alloc>::erase(iterator p, iterator q)
{
  std::copy(q, end(), p);
  size_type erase_count = q - p;
  detail::destroy_n(end() - erase_count, erase_count);
  size_ -= erase_count;
  return p;
}

template <typename T, typename Alloc>
void ArrayVector<T, Alloc>::clear()
{
  detail::destroy_n(data_, size_);
  size_ = 0;
}

template <typename T, typename Alloc>
void ArrayVector<T, Alloc>::reserve(size_type new_capacity)
{
  if (new_capacity <= capacity_) return;
  pointer new_data = reserve_raw(new_capacity);
  if (size_ > 0) std::uninitialized_copy(data_, data_ + size_, new_data);
  deallocate(data_, size_);
  data_ = new_data;
  capacity_ = new_capacity;
}

template <typename T, typename Alloc>
void ArrayVector<T, Alloc>::reserve()
{
  if (capacity_ == 0)
  {
    reserve(minimum_capacity);
  }
  else if (size_ == capacity_)
  {
    reserve(2 * capacity_);
  }
}

template <typename T, typename Alloc>
void ArrayVector<T, Alloc>::resize(size_type new_size, value_type const &initial)
{
  if (new_size < size_)
  {
    erase(begin() + new_size, end());
  }
  else if (size_ < new_size)
  {
    insert(end(), new_size - size(), initial);
  }
}
  
// Swap Method
template <typename T, typename Alloc>
void ArrayVector<T, Alloc>::swap(this_type &rhs)
{
  std::swap(size_, rhs.size_);
  std::swap(capacity_, rhs.capacity_);
  std::swap(data_, rhs.data_);
}

//----------------------------------------------------------------
template <typename T, typename Alloc>
void ArrayVector<T, Alloc>::deallocate(pointer data, size_type size)
{
  if (data)
  {
    detail::destroy_n(data, size);
    alloc_.deallocate(data, size);
  }
}

template <typename T, typename Alloc>
typename ArrayVector<T, Alloc>::pointer ArrayVector<T, Alloc>::reserve_raw(size_type capacity)
{
  pointer data = 0;
  if (capacity)
  {
    data = alloc_.allocate(capacity);
  }
  return data;
}

//----------------------------------------------------------------
#ifdef _MSC_VER
  #pragma warning( pop )
#endif
