#include "stdafx.h"
#include "Replacer.h"
#include "Main.h"

typedef unsigned char Byte; 
typedef ArrayVector<Byte> ByteArrayVector;

static bool LoadBinaryData(std::string const &filename, ByteArrayVector &data)
{
  FILE *fp = NULL;
  errno_t err = ::fopen_s(&fp, filename.c_str(), "rb");
  if (err != 0)
  {
    Printf("Failed to open \"%s\"!\n", filename.c_str());
    return false;
  }

  long fsize = ::_filelength(::_fileno(fp));
  data.resize(fsize);
  if (::fread(data.data(), fsize, 1, fp) != 1)
  {
    Printf("Failed to read from \"%s\"!\n", filename.c_str());
    return false;
  }

  ::fclose(fp);
  return true;
}

static bool SaveBinaryData(std::string const &filename, ByteArrayVector const &data)
{
  FILE *fp = NULL;
  errno_t err = ::fopen_s(&fp, filename.c_str(), "wb");
  if (err != 0)
  {
    Printf("Failed to open \"%s\"!\n", filename.c_str());
    return false;
  }

  if (::fwrite(data.data(), data.size(), 1, fp) != 1)
  {
    Printf("Failed to write to \"%s\"!\n", filename.c_str());
    return false;
  }

  ::fclose(fp);
  return true;
}

// Find target in data.
// @param index start index.
// @return returns found index if succeeded. returns -1 if failed.
int FindByteArray(ByteArrayVector const &data, int index, std::string const &target)
{
  for ( ; index < (int)data.size(); ++index)
  {
    bool matched = true;
    for (int j = 0; matched && j < (int)target.size(); ++j)
    {
      if (data[index + j] != target[j]) matched = false;
    }
    if (matched) return index;
  }
  return -1;
}

// Read reference name.
static int ReadName(ByteArrayVector const &data, int idx, std::string &name, int maxNameSize)
{
  // Read reference name.
  int nameSize = data[idx];
  nameSize = std::min(nameSize, maxNameSize);
  name.resize(nameSize);
  for (int i = 0; i < nameSize; ++i) name[i] = data[idx + i + 1];

  // Check postfix.
  Byte postfix1 = data[idx + nameSize + 1];
  Byte postfix2 = data[idx + nameSize + 2];
  if ((0x80 == postfix2) || (0x81 == postfix2) || (0xff == postfix1 && 0xff == postfix2))
  {
    return idx + nameSize + 3;
  }
  else
  {
    return -1;
  }
}

// Confirm the matching of the Pattern A.
// @param idx  The index of after the placeholder.
static int CheckPatternA(ByteArrayVector const &data, int idx, std::string &name, int maxNameSize)
{
  /*if (0x01 != data[idx]     || 0x00 != data[idx + 1] || 0x00 != data[idx + 2] ||
      0x2B != data[idx + 3] || 0x00 != data[idx + 4] || 0x00 != data[idx + 5] || 0x00 != data[idx + 6])
  {
    return -1;
  }*/

  return ReadName(data, idx + 9, name, maxNameSize);
}

// Confirm the matching of the Pattern B.
// @param idx  The index of after the placeholder.
static int CheckPatternB(ByteArrayVector const &data, int idx, std::string &name, int maxNameSize)
{
  // Search ".psl"
  int idxPSL = FindByteArray(data, idx, ".psl");
  if (idxPSL == -1) return -1;

  return ReadName(data, idxPSL + 6, name, maxNameSize);
}

// Confirm the matching of the Pattern C.
// @param idx  The index of after the placeholder.
static int CheckPatternC(ByteArrayVector const &data, int idx, std::string &name, int maxNameSize)
{
  // Search ".cml"
  int idxCML = FindByteArray(data, idx, ".cml");
  if (idxCML == -1) return -1;

  if (0x00 != data[idxCML + 4] || 0x00 != data[idxCML + 5] || 0x00 != data[idxCML + 6] ||
      0x00 != data[idxCML + 7] || 0x00 != data[idxCML + 8])
  {
    return -1;
  }

  return ReadName(data, idxCML + 9, name, maxNameSize);
}

static bool ReplaceReference(ByteArrayVector &data, std::string const &placeholder)
{
  int idx = 0;

  // Update progress bar.
  SetProgBar(0, 100, 0);

  bool end = false;
  while (!end)
  {
    // Search the placeholder.
    int idxPlaceholder = FindByteArray(data, idx, placeholder);
    if (-1 == idxPlaceholder)
    {
      // Not found
      end = true;
    }
    else
    {
      // Found
      std::string name;
      int newIdx = CheckPatternA(data, idxPlaceholder + placeholder.size(), name, placeholder.size());
      if (newIdx == -1)
      {
        newIdx = CheckPatternB(data, idxPlaceholder + placeholder.size(), name, placeholder.size());
      }
      if (newIdx == -1)
      {
        newIdx = CheckPatternC(data, idxPlaceholder + placeholder.size(), name, placeholder.size());
      }

      if (newIdx == -1)
      {
        Printf("Can't find reference name at 0x%08x\n", idxPlaceholder);

        // Skip this placeholder.
        idx = idxPlaceholder + placeholder.size();
      }
      else
      {
        Printf("Replace %-8s at 0x%08x\n", name.c_str(), idxPlaceholder);
        
        // Copy the reference name.
        for (int i = 0; i < (int)placeholder.size(); ++i)
        {
          if (i < (int)name.size()) data[idxPlaceholder + i] = name[i];
          else data[idxPlaceholder + i] = ' ';
        }
        idx = newIdx;
      }
    }

    // Update progress bar.
    SetProgBar(0, 100, (short)((double)idxPlaceholder * 100.0f / data.size()));
  }

  // Update progress bar.
  SetProgBar(0, 100, 100);

  return true;
}

bool ReplaceRefName(std::string inputFilename, std::string outputFilename)
{
  ByteArrayVector data;
  if (!LoadBinaryData(inputFilename, data))
  {
    return false;
  }

  if (!ReplaceReference(data, "<Placeholder>"))
  {
    return false;
  }

  if (!SaveBinaryData(outputFilename, data))
  {
    return false;
  }
  
  return true;
}
